
/* Copyright (c) Mark Steven Davis (G0KHB), 2002. */

/* ====================================================================== */

#define DEBUG   FALSE

#include <stdlib.h>

#if DEBUG
#include <rep.h>
#endif

#include "listhand.h"

/* ====================================================================== */

#define MAX_LISTS       ( 20 )
#define BAD_LIST        ( -1 )

/* ====================================================================== */

struct  LNodeType
        {
                struct  LNodeType       *Prev;
                struct  LNodeType       *Next;
                int                     Size;
                char                    Data;
        };

struct  LHeadType
        {
                BOOL                    InUse;
                struct  LNodeType       *First;
                struct  LNodeType       *Curr;
                struct  LNodeType       *Last;
        };

typedef struct LNodeType LNode;
typedef struct LHeadType LHead;

/* ====================================================================== */

BOOL    Initialised = FALSE;

LHead   LSets[MAX_LISTS];

/* ====================================================================== */

BOOL    Debug = FALSE;

#if     DEBUG
#define DB(x) if ( Debug ) { x }
#else
#define DB(x)
#endif

/* ====================================================================== */

void    ListInit(       void );

/* ====================================================================== */

#if     defined(__TURBOC__)
#pragma warn -par
#endif

void    HeadDump(       char    *Title,
                        LHead   *h )
{
        DB(reporta(Title,"&h",h);)
        DB(reporta(Title,"h->First",h->First);)
        DB(reporta(Title,"h->Curr",h->Curr);)
        DB(reporta(Title,"h->Last",h->Last);)
}

#if     defined(__TURBOC__)
#pragma warn .par
#endif

/* ---------------------------------------------------------------------- */

#if     defined(__TURBOC__)
#pragma warn -par
#endif

void    NodeDump(       char    *Title,
                        LNode   *n )
{
        DB(reporta(Title,"&n",n);)
        DB(reporta(Title,"n->Prev",n->Prev);)
        DB(reporta(Title,"n->Next",n->Next);)
        DB(reportn(Title,"n->Size",n->Size);)
        DB(reportd(Title,"n->Data",&n->Data,n->Size);)
}

#if     defined(__TURBOC__)
#pragma warn .par
#endif

/* ====================================================================== */

BOOL    NodeAdd(        LHead   *h,
                        int     DataLen,
                        void    *DataPtr )
{
        LNode   *n;

        if ( h == NULL ) return FALSE;

        n = (LNode*) malloc(sizeof(LNode)+DataLen);
        if ( n == NULL ) return FALSE;

        n->Size = DataLen;
        memcpy(&n->Data,DataPtr,DataLen);

        if      ( h->First == NULL )    /* Link in as *only* item. */
        {
                n->Prev = NULL;
                n->Next = NULL;

                h->First = n;
                h->Curr = n;
                h->Last = n;
        }
        else
        if      ( h->Curr == NULL )     /* Link in at *head* of chain. */
        {
                n->Next = h->First;
                n->Next->Prev = n;
                n->Prev = NULL;

                h->First = n;
                h->Curr = n;
        }
        else                            /* Link in after *current* node. */
        {
                n->Next = h->Curr->Next;
                n->Prev = h->Curr;
                if ( n->Next != NULL ) n->Next->Prev = n;
                h->Curr->Next = n;

                if ( h->Last == h->Curr ) h->Last = n;
                h->Curr = n;
        }

        return TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    NodeGet(        LHead   *h,
                        int     *DataLen,
                        void    *DataPtr )
{
        LNode   *n;
        int     Dst, Src, Cpy;

        if ( h == NULL ) return FALSE;

        n = h->Curr;
        if ( n == NULL ) return FALSE;

        /* If the source data will fit into the destination buffer, copy the
           entire data set.  Otherwise, copy as much as we can. */

        Src = n->Size, Dst = *DataLen;
        Cpy = ( Src > Dst ) ? Dst : Src;

        memcpy(DataPtr,&n->Data,Cpy);

        /* Always return the *real* size of the source data not just how much
           we managed to copy into the destination buffer. */

        *DataLen = Src;

        return TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    NodeSet(        LHead   *h,
                        int     DataLen,
                        void    *DataPtr )
{
        LNode   *n;

        if ( h == NULL ) return FALSE;

        n = h->Curr;
        if ( n == NULL ) return FALSE;

        /* For the moment, at least, we only allow an update if the old data
           size is *exactly the same as the *new* data size.  Last, we can
           relax that limitation if a little bit of extra code is written. */

        if ( n->Size != DataLen ) return FALSE;

        memcpy(&n->Data,DataPtr,DataLen);

        return TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    NodeDel(        LHead   *h )
{
        LNode   *p, *c, *n;     /* Prev, current, and next nodes in the list. */

        if ( h == NULL ) return FALSE;
        if ( h->Curr == NULL ) return FALSE;

        c = h->Curr;
        p = c->Prev;
        n = c->Next;

        if ( h->First == c ) h->First = n;
        h->Curr = n;
        if ( h->Last == c ) h->Last = p;

        if ( p != NULL ) p->Next = n;
        if ( n != NULL ) n->Prev = p;

        free(c);

        return  TRUE;
}

/* ====================================================================== */

void    ListInit(       void )
{
        int     i;
        LHead   *p;

        DB(report("ListInit","Starts","");)

        for     ( i=ZERO ; i<MAX_LISTS ; i++ )
        {
                p = LSets + i;

                p->InUse = FALSE;
                p->First = NULL;
                p->Curr  = NULL;
                p->Last  = NULL;
        }

        DB(report("ListInit","Ends","");)

        Initialised = TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    FindHead(       int     Handle,
                        LHead   **h )
{
        int     i;

        if ( ( Handle < ZERO ) || ( Handle >= MAX_LISTS ) ) return FALSE;

        *h = LSets + Handle;

        return  TRUE;
}

/* ====================================================================== */

BOOL    ListCreate(     int     *Handle )
{
        int     i = ZERO;
        LHead   *h;

        DB(report("ListNew","Starts","");)

        if ( !Initialised ) ListInit();

        while   ( ( i < MAX_LISTS ) && LSets[i].InUse ) i++;

        if ( i > MAX_LISTS ) return FALSE;

        if ( !FindHead(i,&h) ) return FALSE;

        h->InUse = TRUE;
        h->First = NULL;
        h->Curr  = NULL;
        h->Last  = NULL;

        DB(reportn("ListNew","New handle",*Handle);)

        *Handle = i;
        return TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    ListAdd(        int     Handle,
                        int     DataLen,
                        void    *DataPtr )
{
        LHead   *h;
        BOOL    Res;

        DB(reportn("ListAdd","Adding to list",Handle);)
        DB(reportd("ListAdd","Data",(char*)DataPtr,DataLen);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        Res = NodeAdd(h,DataLen,DataPtr);

        DB(reportb("ListAdd","Returns",Res);)

        return Res;
}

/* ---------------------------------------------------------------------- */

BOOL    ListFirst(      int     Handle )
{
        LHead   *h;

        DB(reportn("ListFirst","Going to start",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        h->Curr = h->First;

        HeadDump("ListFirst",h);

        return  h->Curr != NULL;
}

/* ---------------------------------------------------------------------- */

BOOL    ListNext(       int     Handle )
{
        LHead   *h;

        DB(reportn("ListNext","Going to next",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        if ( h->Curr != NULL ) h->Curr = h->Curr->Next;

        HeadDump("ListNext",h);

        return  h->Curr != NULL;
}

/* ---------------------------------------------------------------------- */

BOOL    ListGet(        int     Handle,
                        int     *DataLen,
                        void    *DataPtr )
{
        LHead   *h;

        DB(reportn("ListGet","Getting from list",Handle);)
        DB(reportn("ListGet","Max size",*DataLen);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        if ( !NodeGet(h,DataLen,DataPtr) ) return FALSE;

        DB(reportn("ListGet","Size",*DataLen);)
        DB(reportd("ListGet","Data",(char*)DataPtr,*DataLen);)

        return TRUE;
}

/* ---------------------------------------------------------------------- */

BOOL    ListSet(        int     Handle,
                        int     DataLen,
                        void    *DataPtr )
{
        LHead   *h;

        DB(reportn("ListSet","Setting into list",Handle);)
        DB(reportn("ListSet","Size",DataLen);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        return NodeSet(h,DataLen,DataPtr);
}

/* ---------------------------------------------------------------------- */

BOOL    ListPrev(       int     Handle )
{
        LHead   *h;

        DB(reportn("ListPrev","Going to previous",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        if ( h->Curr != NULL ) h->Curr = h->Curr->Prev;

        HeadDump("ListPrev",h);

        return  h->Curr != NULL;
}

/* ---------------------------------------------------------------------- */

BOOL    ListLast(       int     Handle )
{
        LHead   *h;

        DB(reportn("ListLast","Going to end",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        h->Curr = h->Last;

        HeadDump("ListLast",h);

        return  h->Curr != NULL;
}

/* ---------------------------------------------------------------------- */

BOOL    ListDel(        int     Handle )
{
        LHead   *h;
        BOOL    Res;

        DB(reportn("ListDel","Going to delete",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        Res = NodeDel(h);

        DB(reportb("ListDel","Returns",Res);)

        return Res;
}

/* ---------------------------------------------------------------------- */

BOOL    ListDestroy(    int     *Handle )
{
        LHead   *h;

        DB(reportn("ListDestroy","Destroying list",*Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(*Handle,&h) ) return FALSE;
        if ( !h->InUse ) return FALSE;

        h->Curr = h->First;

        while ( NodeDel(h) );

        h->InUse = FALSE;
        *Handle = BAD_LIST;

        return TRUE;
}

/* ---------------------------------------------------------------------- */

#if     defined(__TURBOC__)
#pragma warn -par
#endif

BOOL    ListDump(       int     Handle )
{
#if     DEBUG

        LHead   *h;
        LNode   *n;
        BOOL    OldDebug = Debug;

        Debug = TRUE;

        DB(reportn("ListDump","Dumping list",Handle);)

        if ( !Initialised ) ListInit();

        if ( !FindHead(Handle,&h) ) { Debug = OldDebug; return FALSE; }
        if ( !h->InUse )            { Debug = OldDebug; return FALSE; }

        HeadDump("ListDump",h);

        for     ( n=h->First ; n!=NULL ; n=n->Next )
                NodeDump("ListDump",n);

        DB(report("ListDump","Ends","");)

        Debug = OldDebug;
        return TRUE;
#else
        return FALSE;
#endif
}

#if     defined(__TURBOC__)
#pragma warn .par
#endif

/* ====================================================================== */

